#! /usr/bin/perl -w
#
# @(#)$Id$
#
use strict;
use Socket;
use Getopt::Long qw(:config no_ignore_case bundling);

use vars qw/ $help $verb /;
my $help           = undef;
my $verb           = 0;

GetOptions( "h|help" => \$help, "verbose|v+" => \$verb );

if ( $help ) {
 print STDERR <<EOF;
Usage: $0 [-h] [-v] [ruleset]

  Generate a set of named extended ACLs for Foundry Network devices to be
  applied to the inbound dirction of an interface. The ruleset is expressed
  in bi-directional semantics: based on the list of connected networks to
  each interface a full complement of access control entries is generated
  such that only 'allowed' traffic in permitted to enter the logical routing
  core and thus from there go unfiltered to one of the connected networks.
  This script can only process and generate IPv4 ACLs.

    -h             Give this help
    -v             Give verbose diagnostics on STDERR

Syntax (see also the example file):

   interface <name>
     connects <ipv4-net>
     [connects <ipv4-net> ...]
     [excludes <ipv4-net> ...]
     [prepend <ruledef> ...]
     [append <ruledef> ...]
     end

   stanza <name> [<in-line-replacement>]
     [<next-line-replacement> ...]
     end

   ruleset
     [<rule-in-FN-EACL-syntax> ...]
     end

   end

  The "stanza" lines act as macros and can be invoked by \$macro([args])
  later in the rule set. Any "\$variable" instances are replaced by their 
  value given between the parentheses. Stanzas can be recursive.

EOF
  
  exit(0);
}

sub addr_word($) {
  my ($mask) = @_;
  my ($bmask,$i)=(0,0);
  
  if ( length($mask) <= 2 ) {
    for ($i=0;$i<$mask;$i++) { $bmask|=1<<(31-$i); } 
  } else {
    foreach (split /\./,$mask) { $bmask|=(($_)<<(8*(3-($i++)))); }
  }
  return $bmask;
}

sub netmatch($$) {
  my ($a,$b) = @_;

  my ($ipa,$maska) = split(/\//,$a,2);
  my ($ipb,$maskb) = split(/\//,$b,2);
  
  my $bmaska = &addr_word($maska);
  my $bmaskb = &addr_word($maskb);
  my $shortestmask = $bmaska & $bmaskb;
  
  my $addra=&addr_word($ipa);
  my $addrb=&addr_word($ipb);

  $verb>2 and print "\n  shortest mask: $shortestmask for $ipa $ipb\n";
  $verb>2 and print "  with addra = $addra and addrb = $addrb\n";
  
  return 1 unless (($addra & $shortestmask) ^ ($addrb & $shortestmask));
  return 0;
}

sub netwithin($$) {
  my ($a,$b) = @_;

  my ($ipa,$maska) = split(/\//,$a,2);
  my ($ipb,$maskb) = split(/\//,$b,2);
  
  my $bmaska = &addr_word($maska);
  my $bmaskb = &addr_word($maskb);
  my $shortestmask = $bmaska & $bmaskb;
  
  my $neta=&addr_word($ipa) & $bmaskb & $bmaska;
  my $netb=&addr_word($ipb) & $bmaskb;

  $verb>3 and printf "neta=%04x, netb=%04x, XOR=%04x, maska=%04x, maskb=%04x\n",$neta,$netb,$netb^$neta,$bmaska,$bmaskb; 
  if ( (($netb^$neta)==0) || ($netb==$neta) ) { # they are equal, so may be subset?
    # now depends on netmask length
    my $allones=4294967295;
    if ( ($bmaskb==$bmaska) || (($bmaskb^$allones) & $bmaska) ) { return 1; }
  }
  return 0;
}

sub matchtest($$) {
	my ($a,$b) = @_;
	print "Testing if $a matches $b: ";
	if (&netmatch($a,$b)) { print "  YES"; } else { print "  NO"; }
	print "\n";
}

#&matchtest("127.0.0.0/8","127.0.0.0/8");
#&matchtest("172.16.0.0/12","172.20.0.0/16");
#&matchtest("172.16.0.0/12","172.20.0.0/255.255.0.0");
#&matchtest("192.16.199.0/24","172.20.0.0/16");
#&matchtest("172.21.0.0/16","172.20.0.0/16");

my (%interface,@rules,%stanza,@oldlines);

sub readline() {
	my ($line);
	
	# are there old lines to feed?
	if ( defined $oldlines[0] ) {
		$line = shift @oldlines;
		chomp($line); $line =~ s/^\s+//; 
		$_ = $line;
		return $line;
	}
	
	while (	$line = <> ) { 
	  chomp($line); $line =~ s/^\s+//; 
	  $line =~ /^!/ and next; 
	  $line =~/^$/ and next;
	  last;
	}
	
	if ( defined $line ) {
	 while ( $line =~ m/\$([-\w\d]+)\(([^\)]*)\)/ ) {
	  my ($name,$args) = ($1,$2);
	  die "Stanza $name undefined\n" unless defined $stanza{$name};
	  my $instance = $stanza{$name};
	  foreach my $argdef ( split /,/,$args ) {
	    my ($key,$value) = split /=/,$argdef;
	    $instance =~ s/\$$key([^\(\w])/$value$1/isg;
	  }
	  $line =~ s/\$([-\w\d]+)\(([^\)]*)\)/$instance/;
	 }
	 push @oldlines,split(/\n/,$line);
	}
	return undef if not defined $line;
	
	$line = shift @oldlines;
	
	$_ = $line;
	return $line;
}

sub parse_interface($) {
	my ($name) = @_;
	my (@networks,@excludes);
	
	$interface{$name}{"name"} = $name;
	
	push @networks,"127.0.0.0/8";
	
	while(&readline) {
		/^end/ and do {
			$interface{$name}{"network"} = \@networks;
			$interface{$name}{"excludes"} = \@excludes;
			return 0;
		};
		/^prepend/ and do {
			my ($kw,$rule) = split /\s+/,$_,2;
			push @{$interface{$name}{"prepend"}}, $rule;
		};
		/^append/ and do {
			my ($kw,$rule) = split /\s+/,$_,2;
			push @{$interface{$name}{"append"}}, $rule;
		};
		/^connects/ and do {
			my ($kw,$network) = split;
			push @networks,$network;
		};
		/^excludes/ and do {
			my ($kw,$network) = split;
			push @excludes,$network;
		};
	}
	die "Syntax error in input aroud line $.: \n".
		"no end statement found in interface definition $name\n";
}

sub parse_ruleset() {
	while (&readline) {
		/^end/ and return 0;
		/^(permit|deny)/ and do {
			my ($kw,$rule) = split /\s+/,$_,2;
			my %ruledef = &parse_rule($_);
			push @rules,\%ruledef;
		};
	}
}

sub parse_stanza($) {
	my ($def) = @_;
	my ($name,$value) = split /\s/,$def,2;
	
	$stanza{$name} = $value;
	while(&readline) {
		/^end/ and return 0;
		$stanza{$name}.=$_."\n";
	}
	die "Stanza $name unterminated around line $.\n";
}

sub parse_rule($) {
	my ($str) = @_;
	my (%rule);
	my ($i) = 0;

	my @tok = split /\s+/,$str;
	$rule{"text"} = $str;
	$rule{"sense"} = $tok[$i++];
	$rule{"proto"} = $tok[$i++];
	
	# source IP address
	if ( $tok[$i] eq "any" ) {
		$rule{"src"} = "0.0.0.0/0";
	} elsif ( $tok[$i] eq "host" ) {
		$i++;
		$rule{"src"} = $tok[$i]."/32";
	} elsif ( $tok[$i] =~ /^\d+\.\d+\.\d+\.\d+$/ ) {
		die "Syntax error at line $.: $_\n" 
		  unless $tok[$i+1] =~ /^\d+\.\d+\.\d+\.\d+$/; 
		$rule{"src"} = $tok[$i]."/".$tok[$i+1];
		$i++;
	} elsif ( $tok[$i] =~ /^\d+\.\d+\.\d+\.\d+\/.*/ ) {
		$rule{"src"} = $tok[$i];
	} else {
		die "Syntax error in line $.: $_\n";
	}
	$i++;
	
	# optional port
	if ($tok[$i] =~ /^(eq|gt|lt|neq)$/ ) {
		$i+=2;
	}
	if ($tok[$i] =~ /^established$/ ) {
		$i++;
	}
	if ($tok[$i] =~ /^range$/ ) {
		$i+=3;
	}

	# destination IP address
	if ( $tok[$i] eq "any" ) {
		$rule{"dst"} = "0.0.0.0/0";
	} elsif ( $tok[$i] eq "host" ) {
		$i++;
		$rule{"dst"} = $tok[$i]."/32";
	} elsif ( $tok[$i] =~ /^\d+\.\d+\.\d+\.\d+$/ ) {
		die "Syntax error at line $.: $_\n" 
		  unless $tok[$i+1] =~ /^\d+\.\d+\.\d+\.\d+$/; 
		$rule{"dst"} = $tok[$i]."/".$tok[$i+1];
		$i++;
	} elsif ( $tok[$i] =~ /^\d+\.\d+\.\d+\.\d+\/.*/ ) {
		$rule{"dst"} = $tok[$i];
	} else {
		die "Syntax error in line $.: $_\n";
	}
	$i++;

	return %rule;
}

sub generate_acls() {
	print "!\n! generated ACLs for inbound direction\n!\n";
	foreach my $iface ( sort keys %interface ) {
		print "no ip access-list extended $iface\n";
		print "ip access-list extended $iface\n";
		foreach my $rule (@{$interface{$iface}{"prepend"}} ) {
			print "  $rule\n";
		}

		$verb and do {
		  print "! this interface connects:";
		  foreach my $net ( @{$interface{$iface}{"network"}} ) {
		  	print " $net";
		  }
		  print "\n";
		  print "! except for:";
		  foreach my $net ( @{$interface{$iface}{"excludes"}} ) {
		  	print " $net";
		  }
		  print "\n";
		};
		
		foreach my $rule ( @rules ) {
			# is it applicable?
			my $matched = 0;
			
			foreach my $net (@{$interface{$iface}{"network"}} ) {
			  $verb and 
			    print "  ! matching source ".$rule->{"src"}."\n";
			  &netmatch($net,$rule->{"src"}) and $matched++;
			}

			# permit to self-only is useless, but there's not good algo to keep it out
			my ($nnetworks,$nselfmatches)=(0,0);
			foreach my $net (@{$interface{$iface}{"network"}} ) {
			  $net eq "127.0.0.0/8" and next;
			  $nnetworks++;
			  if ( &netwithin($rule->{"dst"},$net) ) {
			    $nselfmatches++;
			    $verb and 
			      print "  ! from source ".$rule->{"src"}.
			            " to dest ".$rule->{"dst"}.
			            " entirely within $net\n";
			  }
			}
			if ( $nnetworks == $nselfmatches ) {
			    $matched = 0;
			    $verb and 
			      print "  ! killed $nselfmatches: ".
			            $rule->{"text"}."\n";
			}
			
			foreach my $net (@{$interface{$iface}{"excludes"}} ) {
			  $verb and print "  ! $net matching source ".
		                          $rule->{"src"}."\n";
			  if ( &netwithin($rule->{"src"},$net) ) {
			    $matched=0;
			    $verb and print "  ! killed ".$rule->{"text"}."\n";
			    $verb and print "  ! source ".$rule->{"src"}." entirely within $net\n";
			  }
			}
			print "  ".$rule->{"text"}."\n" if $matched;
			$verb and print "  ! suppressed: ".$rule->{"text"}."\n" unless $matched;
		}
		foreach my $rule (@{$interface{$iface}{"append"}} ) {
			print "  $rule\n";
		}
		print "!\n";
	}
}

while (&readline) {
	my ($kw,$name) = split /\s+/,$_,2;
	
	$kw =~ /^stanza/ and &parse_stanza($name);
	$kw =~ /^interface/ and &parse_interface($name);
	$kw =~ /^ruleset/ and &parse_ruleset;
	$kw =~ /^end/ and &generate_acls;
};

0;

	
